"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.SnapshotLegacy = void 0;
const internal_1 = require("@fluidframework/core-utils/internal");
const internal_2 = require("@fluidframework/runtime-utils/internal");
const internal_3 = require("@fluidframework/telemetry-utils/internal");
const constants_js_1 = require("./constants.js");
const properties_js_1 = require("./properties.js");
const snapshotChunks_js_1 = require("./snapshotChunks.js");
/**
 * @internal
 */
class SnapshotLegacy {
    constructor(mergeTree, logger, filename, onCompletion) {
        this.mergeTree = mergeTree;
        this.filename = filename;
        this.onCompletion = onCompletion;
        this.logger = (0, internal_3.createChildLogger)({ logger, namespace: "Snapshot" });
        this.chunkSize =
            mergeTree?.options?.mergeTreeSnapshotChunkSize ?? SnapshotLegacy.sizeOfFirstChunk;
    }
    getSeqLengthSegs(allSegments, approxSequenceLength, startIndex = 0) {
        const segs = [];
        let sequenceLength = 0;
        let segCount = 0;
        let segsWithAttribution = 0;
        while (sequenceLength < approxSequenceLength &&
            startIndex + segCount < allSegments.length) {
            const pseg = allSegments[startIndex + segCount];
            segs.push(pseg);
            if (pseg.attribution) {
                segsWithAttribution++;
            }
            sequenceLength += pseg.cachedLength;
            segCount++;
        }
        (0, internal_1.assert)(segsWithAttribution === 0 || segsWithAttribution === segCount, 0x4bf /* all or no segments should have attribution */);
        const attributionSerializer = this.mergeTree.attributionPolicy?.serializer;
        (0, internal_1.assert)(segsWithAttribution === 0 || attributionSerializer !== undefined, 0x559 /* attribution serializer must be provided when there are segments with attribution. */);
        return {
            version: undefined,
            chunkStartSegmentIndex: startIndex,
            chunkSegmentCount: segCount,
            chunkLengthChars: sequenceLength,
            totalLengthChars: this.header.segmentsTotalLength,
            totalSegmentCount: allSegments.length,
            chunkSequenceNumber: this.header.seq,
            segmentTexts: segs.map((seg) => seg.toJSONObject()),
            attribution: segsWithAttribution > 0 || this.mergeTree.attributionPolicy?.isAttached
                ? attributionSerializer?.serializeAttributionCollections(segs)
                : undefined,
        };
    }
    /**
     * Emits the snapshot to an ISummarizeResult. If provided the optional IFluidSerializer will be used when
     * serializing the summary data rather than JSON.stringify.
     */
    emit(catchUpMsgs, serializer, bind) {
        const chunk1 = this.getSeqLengthSegs(this.segments, this.chunkSize);
        let length = chunk1.chunkLengthChars;
        let segments = chunk1.chunkSegmentCount;
        const builder = new internal_2.SummaryTreeBuilder();
        builder.addBlob(SnapshotLegacy.header, (0, snapshotChunks_js_1.serializeAsMinSupportedVersion)(SnapshotLegacy.header, chunk1, this.logger, this.mergeTree.options, serializer, bind));
        if (chunk1.chunkSegmentCount < chunk1.totalSegmentCount) {
            const chunk2 = this.getSeqLengthSegs(this.segments, this.header.segmentsTotalLength, chunk1.chunkSegmentCount);
            length += chunk2.chunkLengthChars;
            segments += chunk2.chunkSegmentCount;
            builder.addBlob(SnapshotLegacy.body, (0, snapshotChunks_js_1.serializeAsMinSupportedVersion)(SnapshotLegacy.body, chunk2, this.logger, this.mergeTree.options, serializer, bind));
        }
        (0, internal_1.assert)(length === this.header.segmentsTotalLength, 0x05d /* "emit: mismatch in segmentsTotalLength" */);
        (0, internal_1.assert)(segments === chunk1.totalSegmentCount, 0x05e /* "emit: mismatch in totalSegmentCount" */);
        if (catchUpMsgs !== undefined && catchUpMsgs.length > 0) {
            // Messages used to have a "term" property which has since been removed.
            // It is benign so it doesn't really need to be deleted here, but doing so permits snapshot tests
            // to pass with an exact match (and matching the updated definition of ISequencedDocumentMessage).
            for (const message of catchUpMsgs) {
                // eslint-disable-next-line @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-member-access
                delete message.term;
            }
            builder.addBlob(this.mergeTree.options?.catchUpBlobName ?? SnapshotLegacy.catchupOps, serializer ? serializer.stringify(catchUpMsgs, bind) : JSON.stringify(catchUpMsgs));
        }
        return builder.getSummaryTree();
    }
    extractSync() {
        const collabWindow = this.mergeTree.collabWindow;
        this.seq = collabWindow.minSeq;
        this.header = {
            segmentsTotalLength: this.mergeTree.getLength(this.mergeTree.collabWindow.minSeq, constants_js_1.NonCollabClient),
            seq: this.mergeTree.collabWindow.minSeq,
        };
        let originalSegments = 0;
        const segs = [];
        let prev;
        const extractSegment = (segment, pos, refSeq, clientId, start, end) => {
            if (segment.seq !== constants_js_1.UnassignedSequenceNumber &&
                segment.seq <= this.seq &&
                (segment.removedSeq === undefined ||
                    segment.removedSeq === constants_js_1.UnassignedSequenceNumber ||
                    segment.removedSeq > this.seq)) {
                originalSegments += 1;
                if (prev?.canAppend(segment) && (0, properties_js_1.matchProperties)(prev.properties, segment.properties)) {
                    prev = prev.clone();
                    prev.append(segment.clone());
                }
                else {
                    if (prev) {
                        segs.push(prev);
                    }
                    prev = segment;
                }
            }
            return true;
        };
        this.mergeTree.mapRange(extractSegment, this.seq, constants_js_1.NonCollabClient, undefined);
        if (prev) {
            segs.push(prev);
        }
        this.segments = [];
        let totalLength = 0;
        segs.map((segment) => {
            totalLength += segment.cachedLength;
            if (segment.properties !== undefined && Object.keys(segment.properties).length === 0) {
                segment.properties = undefined;
            }
            this.segments.push(segment);
        });
        // To reduce potential spam from this telemetry, we sample only a small
        // percentage of summaries
        if (Math.abs(originalSegments - segs.length) > 500 && Math.random() < 0.005) {
            this.logger.sendTelemetryEvent({
                eventName: "MergeTreeLegacySummarizeSegmentCount",
                originalSegments,
                segmentsAfterCombine: segs.length,
                segmentsLen: this.segments.length,
            });
        }
        // We observed this.header.segmentsTotalLength < totalLength to happen in some cases
        // When this condition happens, we might not write out all segments in getSeqLengthSegs()
        // when writing out "body". Issue #1995 tracks following up on the core of the problem.
        // In the meantime, this code makes sure we will write out all segments properly
        if (this.header.segmentsTotalLength !== totalLength) {
            this.logger.sendErrorEvent({
                eventName: "SegmentsTotalLengthMismatch",
                totalLength,
                segmentsTotalLength: this.header.segmentsTotalLength,
            });
            this.header.segmentsTotalLength = totalLength;
        }
        return this.segments;
    }
}
exports.SnapshotLegacy = SnapshotLegacy;
SnapshotLegacy.header = "header";
SnapshotLegacy.body = "body";
SnapshotLegacy.catchupOps = "catchupOps";
// Split snapshot into two entries - headers (small) and body (overflow) for faster loading initial content
// Please note that this number has no direct relationship to anything other than size of raw text (characters).
// As we produce json for the blob (and then send over the wire compressed), this number
// is really hard to correlate with any actual metric that matters (like bytes over the wire).
// For test with small number of chunks it would be closer to blob size,
// for very chunky text, blob size can easily be 4x-8x of that number.
SnapshotLegacy.sizeOfFirstChunk = 10000;
//# sourceMappingURL=snapshotlegacy.js.map